/*
************************************************************************
*
*   POVText.c - POV text drawing
*
*   Copyright (c) 1994
*
*   ETH Zuerich
*   Institut fuer Molekularbiologie und Biophysik
*   ETH-Hoenggerberg
*   CH-8093 Zuerich
*
*   SPECTROSPIN AG
*   Industriestr. 26
*   CH-8117 Faellanden
*
*   All Rights Reserved
*
*   Date of last modification : 96/03/02
*   Pathname of SCCS file     : /sgiext/molmol/sg/src/pov/SCCS/s.POVText.c
*   SCCS identification       : 1.12
*
************************************************************************
*/

#include "pov_text.h"

#include <string.h>

#include <sg_map.h>
#include <sg_plot.h>
#include "pov_util.h"
#include "pov_color.h"
#include "pov_clip.h"

#define CHAR_W 1.4  /* width in definiton file */
#define CHAR_H 2.0  /* height in definiton file */
#define CHAR_S 0.0  /* spacing */

#define SUP_SCALE 0.65
#define SUP_OFFS 0.4
#define SUB_OFFS -0.25

static float CurrSize2D, CurrSize;

static char *
getCharName(char ch)
{
  static char name[20];

  (void) strcpy(name, "font_");

  if (ch >= 'A' && ch <= 'Z') {
    name[5] = ch;
    name[6] = '\0';
  } else if (ch >= 'a' && ch <= 'z') {
    name[5] = ch;
    name[6] = '\0';
  } else if (ch >= '0' && ch <= '9') {
    name[5] = ch;
    name[6] = '\0';
  } else if (ch == '!') {
    (void) strcpy(name + 5, "exclamation");
  } else if (ch == '"') {
    (void) strcpy(name + 5, "quote");
  } else if (ch == '#') {
    (void) strcpy(name + 5, "pound");
  } else if (ch == '$') {
    (void) strcpy(name + 5, "dollar");
  } else if (ch == '%') {
    (void) strcpy(name + 5, "percent");
  } else if (ch == '\'') {
    (void) strcpy(name + 5, "apostrophe");
  } else if (ch == '(') {
    (void) strcpy(name + 5, "leftp");
  } else if (ch == ')') {
    (void) strcpy(name + 5, "rightp");
  } else if (ch == '*') {
    (void) strcpy(name + 5, "star");
  } else if (ch == '+') {
    (void) strcpy(name + 5, "plus");
  } else if (ch == ',') {
    (void) strcpy(name + 5, "comma");
  } else if (ch == '-') {
    (void) strcpy(name + 5, "minus");
  } else if (ch == '.') {
    (void) strcpy(name + 5, "point");
  } else if (ch == '/') {
    (void) strcpy(name + 5, "slash");
  } else if (ch == ':') {
    (void) strcpy(name + 5, "colon");
  } else if (ch == ';') {
    (void) strcpy(name + 5, "semicolon");
  } else if (ch == '<') {
    (void) strcpy(name + 5, "less");
  } else if (ch == '=') {
    (void) strcpy(name + 5, "equal");
  } else if (ch == '>') {
    (void) strcpy(name + 5, "greater");
  } else if (ch == '[') {
    (void) strcpy(name + 5, "leftp");
  } else if (ch == ']') {
    (void) strcpy(name + 5, "rightp");
  } else if (ch == '^') {
    (void) strcpy(name + 5, "up");
  } else {
    return NULL;
  }

  return name;
}

static void
plotText(float x[3], float fullSize, char *str)
{
  char *chName;
  float currSize, offs, pos;
  BOOL inPar;

  currSize = fullSize;
  offs = 0.0;
  pos = 0.0;
  inPar = FALSE;

  while (*str != '\0') {
    if (*str == '_') {
      str++;
      if (*str == '{') {
	inPar = TRUE;
	str++;
      }
      currSize = SUP_SCALE * fullSize;
      offs = SUB_OFFS * CHAR_H * fullSize;
    } else if (*str == '^') {
      str++;
      if (*str == '{') {
	inPar = TRUE;
	str++;
      }
      currSize = SUP_SCALE * fullSize;
      offs = SUP_OFFS * CHAR_H * fullSize;
    } else {
      if (*str == '}') {
	inPar = FALSE;
	str++;
	continue;
      }
      if (! inPar) {
	currSize = fullSize;
	offs = 0.0;
      }
    }

    if (*str == '\\')
      str++;

    chName = getCharName(*str);
    if (chName != NULL) {
      SgPlotOutputStr("object {\n");
      SgPlotOutputStr(chName);
      SgPlotOutputStr("\n");
      SgPlotOutputStr("scale ");
      SgPlotOutputFloat(currSize);
      SgPlotOutputStr("\n");
      SgPlotOutputStr("rotate <-90, 0, 0>\n");
      SgPlotOutputStr("translate <");
      SgPlotOutputFloat(x[0] + pos);
      SgPlotOutputStr(", ");
      SgPlotOutputFloat(x[1] + offs);
      SgPlotOutputStr(", ");
      SgPlotOutputFloat(x[2]);
      SgPlotOutputStr(">\n");
      POVPlotColor();
      (void) POVClip(x, 0.0);
      SgPlotOutputStr("}\n");
    }

    pos += (CHAR_W + CHAR_S) * currSize;
    str++;
  }
}

void
SgPOVDrawAnnot2D(float x[2], char *str)
{
  float xm[3];

  POVUpdateColor();
  POVMap2D(xm, x);
  plotText(xm, CurrSize2D, str);
}

void
SgPOVDrawText2D(float x[2], char *str)
{
  float xm[3];

  POVUpdateColor();
  POVMap2D(xm, x);
  plotText(xm, CurrSize2D, str);
}

void
SgPOVDrawAnnot(float x[3], char *str)
{
  float xm[3];

  POVUpdateColor();
  SgMapModelPoint(xm, x);
  plotText(xm, CurrSize, str);
}

void
SgPOVDrawText(float x[3], char *str)
{
  float xm[3];

  POVUpdateColor();
  SgMapModelPoint(xm, x);
  plotText(xm, CurrSize, str);
}

static float
mapSize(float size)
{
  float xm[3], x[2];

  x[0] = 0.0;
  x[1] = size;

  POVMap2D(xm, x);

  return xm[1] / CHAR_H;
}

void
SgPOVSetFontSize2D(float size)
{
  CurrSize2D = mapSize(size);
}

void
SgPOVSetFontSize(float size)
{
  CurrSize = mapSize(size);
}
