/*
************************************************************************
*
*   ExCalcSurf.c - CalcSurf command
*
*   Copyright (c) 1994-95
*
*   ETH Zuerich
*   Institut fuer Molekularbiologie und Biophysik
*   ETH-Hoenggerberg
*   CH-8093 Zuerich
*
*   SPECTROSPIN AG
*   Industriestr. 26
*   CH-8117 Faellanden
*
*   All Rights Reserved
*
*   Date of last modification : 96/04/21
*   Pathname of SCCS file     : /sgiext/molmol/src/cmdcalc/SCCS/s.ExCalcSurf.c
*   SCCS identification       : 1.8
*
************************************************************************
*/

#include <cmd_calc.h>

#include <stdio.h>
#include <stdlib.h>
#include <values.h>
#include <math.h>

#include <break.h>
#include <bool.h>
#include <mat_vec.h>
#include <sphere.h>
#include <grid.h>
#include <pu.h>
#include <arg.h>
#include <data_hand.h>
#include <data_sel.h>

typedef struct {
  float partSurf, fullSurf, perc;
  DSTR name;
} ResData;

typedef struct {
  PropRefP refP;
  char *formRes;
  ResData *resDataA;
  float *molSurfA;
  DhAtomP atomP;
  float rad;
  DhAtomP *neighA;
  int neighSize, neighNo;
  int molI;
  GRID grid;
  float rMax;
} SurfData;

static float SolventRad = 1.4;
static int SurfPrec = 3;

static int SpherePointNo;
static Vec3 *SpherePoints;

static int MolNo, MinResI, MaxResI;

static PuTextWindow TextW;

static void
countRes(DhResP resP, void *clientData)
{
  int num;

  num = DhResGetNumber(resP);
  if (num < MinResI)
    MinResI = num;
  if (num > MaxResI)
    MaxResI = num;
}

static void
prepareAtom(DhAtomP atomP, void *clientData)
{
  SurfData *dataP = clientData;
  float rad;
  Vec3 x;

  if (BreakCheck(10000))
    return;

  rad = DhAtomGetVdw(atomP);
  if (rad < 0.0)
    return;

  DhAtomGetCoord(atomP, x);
  GridPrepareAddEntry(dataP->grid, x);

  if (rad > dataP->rMax)
    dataP->rMax = rad;
}

static void
insertAtom(DhAtomP atomP, void *clientData)
{
  SurfData *dataP = clientData;
  float rad;
  Vec3 x;

  if (BreakCheck(10000))
    return;

  rad = DhAtomGetVdw(atomP);
  if (rad < 0.0)
    return;

  DhAtomGetCoord(atomP, x);
  GridInsertEntry(dataP->grid, x, atomP);
}

static BOOL
fillNeigh(void *entryP, Vec3 cent, void *clientData)
{
  DhAtomP nAtomP = entryP;
  SurfData *dataP = clientData;
  Vec3 dx;
  float rs;

  if (nAtomP == dataP->atomP)
    return TRUE;

  DhAtomGetCoord(nAtomP, dx);
  Vec3Sub(dx, cent);
  rs = dataP->rad + DhAtomGetVdw(nAtomP) + SolventRad;

  if (dx[0] * dx[0] + dx[1] * dx[1] + dx[2] * dx[2] > rs * rs)
    return TRUE;

  if (dataP->neighNo == dataP->neighSize) {
    dataP->neighSize *= 2;
    dataP->neighA = realloc(dataP->neighA,
	dataP->neighSize * sizeof(*dataP->neighA));
  }

  dataP->neighA[dataP->neighNo++] = nAtomP;

  return TRUE;
}

static void
surfAtom(DhAtomP atomP, void *clientData)
{
  SurfData *dataP = clientData;
  DhResP resP;
  DhAtomP nAtomP;
  int fullNo, partNo;
  BOOL fullInside, partInside;
  Vec3 cent, nCent, x, dx;
  float nRad;
  int pointI, atomI, ind;

  dataP->rad = DhAtomGetVdw(atomP);
  if (dataP->rad < 0.0)
    return;

  if (BreakCheck(10000 / SpherePointNo))
    return;

  resP = DhAtomGetRes(atomP);
  DhAtomGetCoord(atomP, cent);

  dataP->atomP = atomP;
  dataP->rad += SolventRad;
  dataP->neighNo = 0;

  GridFind(dataP->grid, cent, 1, fillNeigh, dataP);

  fullNo = 0;
  partNo = 0;

  for (pointI = 0; pointI < SpherePointNo; pointI++) {
    Vec3Copy(x, cent);
    Vec3ScaleAdd(x, dataP->rad, SpherePoints[pointI]);

    fullInside = FALSE;
    partInside = FALSE;

    for (atomI = 0; atomI < dataP->neighNo; atomI++) {
      nAtomP = dataP->neighA[atomI];

      DhAtomGetCoord(nAtomP, nCent);
      Vec3Copy(dx, x);
      Vec3Sub(dx, nCent);

      nRad = DhAtomGetVdw(nAtomP) + SolventRad;
      if (dx[0] * dx[0] + dx[1] * dx[1] + dx[2] * dx[2] > nRad * nRad)
	continue;
      
      fullInside = TRUE;

      if (resP == DhAtomGetRes(nAtomP)) {
	partInside = TRUE;
	break;
      }
    }

    if (! fullInside)
      fullNo++;

    if (! partInside)
      partNo++;
  }

  ind = (DhResGetNumber(resP) - MinResI) * MolNo + dataP->molI;
  dataP->resDataA[ind].fullSurf += fullNo * dataP->rad * dataP->rad;
  dataP->resDataA[ind].partSurf += partNo * dataP->rad * dataP->rad;
}

static void
surfMol(DhMolP molP, void *clientData)
{
  SurfData *dataP = clientData;

  if (BreakInterrupted())
    return;

  dataP->grid = GridNew();
  dataP->rMax = - MAXFLOAT;
  DhMolApplyAtom(dataP->refP, molP, prepareAtom, dataP);
  if (dataP->rMax > 0.0) {
    GridInsertInit(dataP->grid, 2.0 * (dataP->rMax + SolventRad));
    DhMolApplyAtom(dataP->refP, molP, insertAtom, dataP);
    DhMolApplyAtom(dataP->refP, molP, surfAtom, dataP);
  } else {
    /* must be initialized, otherwise GridDestroy crashes */
    GridInsertInit(dataP->grid, 10.0);
  }

  GridDestroy(dataP->grid);

  dataP->molI++;
}

static char *
getForm(int maxVal)
{
  if (maxVal >= 100000)
    return "%6d";

  if (maxVal >= 10000)
    return "%5d";

  if (maxVal >= 1000)
    return "%4d";

  if (maxVal >= 100)
    return "%3d";

  if (maxVal >= 10)
    return "%2d";

  return "%1d";
}

static void
writeMol(DhMolP molP, void *clientData)
{
  SurfData *dataP = clientData;
  DhResP resP;
  DSTR str;
  char buf[20];
  int resI, ind;

  if (BreakInterrupted())
    return;

  PuWriteStr(TextW, "\n");
  PuWriteStr(TextW, DStrToStr(DhMolGetName(molP)));
  PuWriteStr(TextW, ":\n\n");

  str = DStrNew();

  dataP->molSurfA[dataP->molI] = 0.0;

  for (resI = MinResI; resI <= MaxResI; resI++) {
    if (BreakCheck(100))
      break;

    resP = DhResFind(molP, resI);
    if (resP == NULL)
      continue;

    ind = (resI - MinResI) * MolNo + dataP->molI;
    if (dataP->resDataA[ind].partSurf == 0.0)
      dataP->resDataA[ind].perc = 0.0;
    else
      dataP->resDataA[ind].perc = 100.0 * dataP->resDataA[ind].fullSurf /
	  dataP->resDataA[ind].partSurf;

    (void) sprintf(buf, dataP->formRes, resI);
    DStrAssignStr(str, buf);
    (void) sprintf(buf, " %-5s", DStrToStr(DhResGetName(resP)));
    DStrAppStr(str, buf);
    (void) sprintf(buf, " %5.1f", dataP->resDataA[ind].fullSurf);
    DStrAppStr(str, buf);
    (void) sprintf(buf, ", %5.1f%%", dataP->resDataA[ind].perc);
    DStrAppStr(str, buf);
    (void) sprintf(buf, " of %5.1f", dataP->resDataA[ind].partSurf);
    DStrAppStr(str, buf);
    DStrAppStr(str, "\n");

    PuWriteStr(TextW, DStrToStr(str));

    dataP->molSurfA[dataP->molI] += dataP->resDataA[ind].fullSurf;

    dataP->resDataA[ind].name = DhResGetName(resP);
  }

  DStrAssignStr(str, "\nTotal surface: ");
  (void) sprintf(buf, "%8.1f", dataP->molSurfA[dataP->molI]);
  DStrAppStr(str, buf);
  DStrAppStr(str, "\n");
  PuWriteStr(TextW, DStrToStr(str));

  DStrFree(str);

  dataP->molI++;
}

#define ARG_NUM 2

ErrCode
ExCalcSurface(char *cmd)
{
  ArgDescr arg[ARG_NUM];
  ErrCode errCode;
  PropRefP refP;
  SurfData data;
  int resNo, resI;
  ResData minData, maxData, avgData;
  float minSurf, maxSurf, avgSurf;
  DSTR str, resName;
  char buf[20];
  int validNo, molI, ind, i;

  arg[0].type = AT_DOUBLE;
  arg[1].type = AT_INT;

  ArgInit(arg, ARG_NUM);

  arg[0].prompt = "Solvent Radius";
  arg[0].v.doubleVal = SolventRad;

  arg[1].prompt = "Precision";
  arg[1].v.intVal = SurfPrec;

  errCode = ArgGet(arg, ARG_NUM);
  if (errCode != EC_OK) {
    ArgCleanup(arg, ARG_NUM);
    return errCode;
  }

  SolventRad = arg[0].v.doubleVal;
  SurfPrec = arg[1].v.intVal;

  ArgCleanup(arg, ARG_NUM);

  DhActivateGroups(TRUE);

  MolNo = SelMolGet(NULL, 0);

  refP = PropGetRef(PROP_SELECTED, FALSE);

  MinResI = MAXINT;
  MaxResI = - MAXINT;
  DhApplyRes(refP, countRes, NULL);
  if (MinResI > MaxResI) {
    DhActivateGroups(FALSE);
    CipSetError("at least one residue must be selected");
    return EC_ERROR;
  }
  resNo = MaxResI - MinResI + 1;

  BreakActivate(TRUE);

  TextW = PuCreateTextWindow(cmd);
  PuWriteStr(TextW, "Solvent accessible surface:\n");

  SphereCalcPoints(SurfPrec, &SpherePoints, &SpherePointNo);

  data.resDataA = malloc(MolNo * resNo * sizeof(*data.resDataA));
  for (i = 0; i < MolNo * resNo; i++) {
    data.resDataA[i].fullSurf = 0.0;
    data.resDataA[i].partSurf = 0.0;
    data.resDataA[i].name = NULL;
  }
  data.molSurfA = malloc(MolNo * sizeof(*data.molSurfA));

  data.refP = refP;
  data.neighSize = 5;
  data.neighA = malloc(data.neighSize * sizeof(data.neighA));

  data.molI = 0;
  DhApplyMol(refP, surfMol, &data);

  free(data.neighA);
  free(SpherePoints);

  for (i = 0; i < MolNo * resNo; i++) {
    data.resDataA[i].fullSurf *= 4.0 * M_PI / SpherePointNo;
    data.resDataA[i].partSurf *= 4.0 * M_PI / SpherePointNo;
  }

  data.formRes = getForm(MaxResI);

  data.molI = 0;
  DhApplyMol(refP, writeMol, &data);

  PuWriteStr(TextW, "\nSummary:\n\n");
  str = DStrNew();

  for (resI = MinResI; resI <= MaxResI; resI++) {
    if (BreakCheck(100))
      break;

    minData.fullSurf = MAXFLOAT;
    minData.partSurf = MAXFLOAT;
    minData.perc = 100.0;

    maxData.fullSurf = 0.0;
    maxData.partSurf = 0.0;
    maxData.perc = 0.0;

    avgData.fullSurf = 0.0;
    avgData.partSurf = 0.0;
    avgData.perc = 0.0;
    validNo = 0;

    resName = NULL;

    for (molI = 0; molI < MolNo; molI++) {
      ind = (resI - MinResI) * MolNo + molI;

      if (data.resDataA[ind].name == NULL)
	continue;

      if (data.resDataA[ind].fullSurf < minData.fullSurf)
	minData.fullSurf = data.resDataA[ind].fullSurf;
      if (data.resDataA[ind].partSurf < minData.partSurf)
	minData.partSurf = data.resDataA[ind].partSurf;
      if (data.resDataA[ind].perc < minData.perc)
	minData.perc = data.resDataA[ind].perc;

      if (data.resDataA[ind].fullSurf > maxData.fullSurf)
	maxData.fullSurf = data.resDataA[ind].fullSurf;
      if (data.resDataA[ind].partSurf > maxData.partSurf)
	maxData.partSurf = data.resDataA[ind].partSurf;
      if (data.resDataA[ind].perc > maxData.perc)
	maxData.perc = data.resDataA[ind].perc;
      
      avgData.fullSurf += data.resDataA[ind].fullSurf;
      avgData.partSurf += data.resDataA[ind].partSurf;
      avgData.perc += data.resDataA[ind].perc;

      if (resName == NULL)
	resName = data.resDataA[ind].name;

      validNo++;
    }

    if (validNo == 0)
      continue;

    (void) sprintf(buf, data.formRes, resI);
    DStrAssignStr(str, buf);
    (void) sprintf(buf, " %-5s", DStrToStr(resName));
    DStrAppStr(str, buf);
    (void) sprintf(buf, " %5.1f", avgData.fullSurf / validNo);
    DStrAppStr(str, buf);
    (void) sprintf(buf, " (%5.1f", minData.fullSurf);
    DStrAppStr(str, buf);
    (void) sprintf(buf, "..%5.1f)", maxData.fullSurf);
    DStrAppStr(str, buf);
    (void) sprintf(buf, ", %5.1f%%", avgData.perc / validNo);
    DStrAppStr(str, buf);
    (void) sprintf(buf, " (%5.1f", minData.perc);
    DStrAppStr(str, buf);
    (void) sprintf(buf, "..%5.1f)", maxData.perc);
    DStrAppStr(str, buf);
    (void) sprintf(buf, " of %5.1f", avgData.partSurf / validNo);
    DStrAppStr(str, buf);
    (void) sprintf(buf, " (%5.1f", minData.partSurf);
    DStrAppStr(str, buf);
    (void) sprintf(buf, "..%5.1f)", maxData.partSurf);
    DStrAppStr(str, buf);
    DStrAppStr(str, "\n");

    PuWriteStr(TextW, DStrToStr(str));
  }

  BreakActivate(FALSE);

  minSurf = MAXFLOAT;
  maxSurf = 0.0;
  avgSurf = 0.0;

  for (molI = 0; molI < MolNo; molI++) {
    if (data.molSurfA[molI] < minSurf)
      minSurf = data.molSurfA[molI];
    if (data.molSurfA[molI] > maxSurf)
      maxSurf = data.molSurfA[molI];
    avgSurf += data.molSurfA[molI];
  }
  avgSurf /= MolNo;

  DStrAssignStr(str, "\nTotal surface: ");
  (void) sprintf(buf, "%8.1f", avgSurf);
  DStrAppStr(str, buf);
  (void) sprintf(buf, " (%8.1f", minSurf);
  DStrAppStr(str, buf);
  (void) sprintf(buf, "..%8.1f)", maxSurf);
  DStrAppStr(str, buf);
  DStrAppStr(str, "\n");
  PuWriteStr(TextW, DStrToStr(str));

  DStrFree(str);
  free(data.resDataA);
  free(data.molSurfA);

  DhActivateGroups(FALSE);

  return EC_OK;
}
